@@ -3,6 +3,7 @@
 #include <string>
 #include <map>
 #include <atomic>
+#include <mutex>
 #include <list>
 #include <vector>
 
@@ -21,7 +22,9 @@ enum class CellType : unsigned char {
     T_STRING_INLINE = 4,
     T_BOOLEAN = 5,
     T_ERROR = 6,
-    T_DATE = 7
+    T_DATE = 7,
+
+    T_SKIP = 8
 };
 
 struct LocationInfo {
@@ -37,10 +40,27 @@ struct XlsxCell {
         unsigned long long integer;
         bool boolean;
     } data;
+    CellType type;
+
+    XlsxCell() : type(CellType::T_NONE) {}
 };
 
 class XlsxFile;
 class XlsxSheet {
+private:
+    template<std::size_t num_buffers>
+    struct ParseState {
+        const size_t threadId;
+        std::array<unsigned char*, num_buffers>& buffers;
+        const size_t bufferSize;
+        const std::atomic_size_t& writeIndex;
+        const std::atomic_bool& finishedWriting;
+        std::vector<std::atomic_size_t>& readIndexes;
+        std::atomic_bool& terminate;
+        std::pair<unsigned long, unsigned long> maxCell;
+        std::mutex& headerMutex;
+        std::atomic_int& headerDone;
+    };
 public:
     unsigned long mSkipRows = 0;
     unsigned long mSkipColumns = 0;
@@ -50,7 +70,7 @@ public:
     mz_zip_archive* mFile;
     int mArchiveIndex;
 
-    std::vector<std::list<std::pair<std::vector<XlsxCell>, std::vector<CellType>>>> mCells;
+    std::vector<std::list<std::vector<XlsxCell>>> mCells;
     std::vector<std::vector<LocationInfo>> mLocationInfos;
 
     bool mHeaders;
@@ -59,8 +79,27 @@ public:
     XlsxSheet(XlsxSheet&& sheet);
 
     XlsxSheet(XlsxFile& parentFile, mz_zip_archive* file, int archiveIndex);
+
+    void specifyTypes(std::vector<CellType> colTypesByIndex, std::map<std::string, CellType> colTypesByName);
     
     bool interleaved(const int skipRows, const int skipColumns, const int numThreads);
     template<std::size_t num_buffers>
-    void interleavedFunc(size_t numThreads, const size_t threadId, std::array<unsigned char*, num_buffers>& buffers, const size_t bufferSize, const std::atomic_size_t& writeIndex, const std::atomic_bool& finishedWriting, std::vector<std::atomic_size_t>& readIndexes, const std::atomic_bool& terminate);
+    void interleavedFunc(size_t numThreads, ParseState<num_buffers>& parseState);
+
+    // for iteration
+    std::pair<size_t, std::vector<XlsxCell>> nextRow();
+    
+    std::vector<CellType> colTypesByIndex;
+private:
+    bool specifiedTypes;
+    std::map<std::string, CellType> colTypesByName;
+
+    // iteration state data
+    size_t maxBuffers;
+    size_t currentBuffer;
+    size_t currentThread;
+    size_t currentCell;
+    unsigned long currentColumn;
+    long long currentRow;
+    std::vector<size_t> currentLocs;
 };
